//  curses based ping and traceroute
//  Willem A Scheuder
//  willem@prinmath.com
//  June 25, 2018 - version 1
//  July 21, 2020 - version 2
//
//  This code is released under the GNU Public License Version 2.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <ncurses.h>
#include <pthread.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netdb.h> 
//  Buttons for GPIO (BCM names)
#ifdef piGPIO
#include <pigpio.h>
int SW[] = {27,23,22,17};
double swt=0;
#endif
//  Provide missing definitions for OSX
#ifdef __APPLE__
   #define SOL_IP IPPROTO_IP
   #define ICMP_TIME_EXCEEDED ICMP_TIMXCEED
   struct icmphdr
   {
      u_int8_t  type;
      u_int8_t  code;
      u_int16_t checksum;
      union
      {
         struct
         {
            u_int16_t id;
            u_int16_t sequence;
         } echo;
         u_int32_t gateway;
      } un;
   };
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
extern char *optarg;
extern int  optind,opterr;

//  Help text
char* help =
   "  ^    Scroll up\n"
   "  v    Scroll down\n"
   "  <    Select previous router\n"
   "  >    Select next router\n"
   "ENTER  Traceroute to router\n"
   " ESC   Return to ping screen\n"
   "  i    Invert colors\n"
   "  r    Reverse direction\n"
   "  t    Toggle time statistics\n"
   "  s    Toggle sound\n"
   "  a    Toggle address\n"
   "  n    Toggle hop count\n"
   "  c    Toggle character\n"
#ifdef piGPIO
   "  g    Enable pi GPIO access\n"
#endif
   "  h    Help\n"
   "  q    Quit program\n";

//  Ping TTL
#define pTTL 64
//  Trace TTL
#define tTTL 24
//  Max length of ping trace in seconds
#define nsec 512

typedef struct
{
   in_addr_t ip;    // IP address
   char*     fqdn;  // Display name
   char*     addr;  // IP address text
} DNS;
typedef struct
{
   in_addr_t ip;         // IP address
   char*     fqdn;       // FQDN
   char*     addr;       // IP address text
   double    dt;         // milliseconds
   double    min;        // Minimum ping
   double    avg;        // Average ping
   double    max;        // Maximum ping
   uint8_t   ping[nsec]; // Ping replies
} Trace;
typedef struct
{
   char*           hdr;        // Header
   char*           name;       // Display name
   char*           host;       // Hostname or IP
   uint8_t         ping[nsec]; // Ping replies
   double          dt;         // milliseconds
   double          min;        // Minimum ping
   double          avg;        // Average ping
   double          max;        // Maximum ping
   int             ttl;        // TTL
   struct in_addr  ip;         // IP address
   struct sockaddr sa;         // Socket address
} Target;

int     mode=0;       //  Mode 1=traceroute, 0=ping, -1=help
int     white=1;      //  White background
int     sbp=1;        //  Seconds between ping
int     r2l=1;        //  Right to left
int     ntar;         //  Number of targets
int     nhdr;         //  Number of header lines
int     sel=0;        //  Selected target
Target* pt;           //  Ping target
int     sock;         //  ICMP socket
int     pingid;       //  PID to identify ping packets
int     seq;          //  Ping sequence number
int     wid;          //  Window width
int     hgt;          //  Window height
int     top=0;        //  Top entry in display
int     nping;        //  Number of pings shown
int     nwid;         //  Name width
int     awid;         //  Addres width
pthread_t rd;         //  Read thread
pthread_t wr;         //  Write thread
int     show=1;       //  Update display
Trace*  tt;           //  Traceroute array
int     tseq;         //  Trace sequence number
int     hop=1;        //  Show hops with ping table
int     nhop=0;       //  Number of traceroute hops
int     stat=0;       //  Display ping time stats
int     pstat=0;      //  Number of pings
int     tstat=0;      //  Number of traces
int     traceid;      //  PID to identify ping packets
DNS*    dns;          //  DNS
int     ndns;         //  Number of DNS entries
int     maxdns;       //  Maximum DNS entries
int     silent=0;     //  Do not beep
int     showip=0;     //  Show IP address
int     pus=1000;     //  Time between pings (in uS)
char    pch=0;        //  Ping character
int     ich=0;        //  Ping symbol
int     swx=0;        //  Switch

//
//  Current time (double)
//
void timeprint(void)
{
   time_t t =  time(NULL);
   struct tm*  l = localtime(&t);
   printw("%4d-%.2d-%.2d %.2d:%.2d:%.2d   #%d  Period %ds Ping time",l->tm_year+1900,l->tm_mon+1,l->tm_mday,l->tm_hour,l->tm_min,l->tm_sec,seq,sbp);
   if (ich==3)
   {
      attron(A_BOLD);
      attron(COLOR_PAIR(2));
      printw(" x1");
      attron(COLOR_PAIR(3));
      printw(" x10");
      attron(COLOR_PAIR(4));
      printw(" x100");
      attron(COLOR_PAIR(5));
      printw(" x1000");
      attroff(A_BOLD);
   }
   else
   {
      attron(COLOR_PAIR(2));
      printw(" <10");
      attron(COLOR_PAIR(3));
      printw(" 10-99");
      attron(COLOR_PAIR(4));
      printw(" 100-999");
      attron(COLOR_PAIR(5));
      printw(" >1000");
   }
   if (silent)
   {
      attron(COLOR_PAIR(5));
      printw(" SILENT");
   }
   attron(COLOR_PAIR(1));
   printw("\n");
}

//
//  Current time (double)
//
double now(void)
{
   struct timeval tv;
   gettimeofday(&tv,(struct timezone *)NULL);
   return tv.tv_sec + ((double)tv.tv_usec)/1000000;
}

//
//  Print error and exit
//
void Fatal(const char* format , ...)
{
   va_list args;
   va_start(args,format);
   vfprintf(stderr,format,args);
   va_end(args);
   exit(1);
}
void FatalW(const char* format , ...)
{
   va_list args;
   endwin();
   va_start(args,format);
   vfprintf(stderr,format,args);
   va_end(args);
   exit(1);
}

//
//  Initialize DNS cache structure
//
void InitDNS()
{
   maxdns = 1024;
   dns = (DNS*)malloc(maxdns*sizeof(DNS));
   if (!dns) FatalW("Cannot allocate DNS memory");
   ndns = 1;
   dns[0].ip = 0;
   dns[0].fqdn = "*";
   dns[0].addr = "*";
}

//
//  Look up DNS address with local cache
//
int nslookup(Trace* tr)
{
   //  Look up IP address
   for (int k=0;k<ndns;k++)
      if (tr->ip == dns[k].ip)
      {
         tr->addr = dns[k].addr;
         tr->fqdn = dns[k].fqdn;
         return strlen(dns[k].fqdn);
      }
   //  Increase array if necessary
   int k = ndns++;
   if (ndns>maxdns)
   {
      maxdns += 1024;
      dns = (DNS*)realloc(dns,maxdns*sizeof(DNS));
      if (!dns) FatalW("Cannot allocate DNS memory");
   }
   //  Set IP
   dns[k].ip = tr->ip;
   char addr[16];
   inet_ntop(AF_INET,&tr->ip,addr,16);
   tr->addr = dns[k].addr = malloc(16);
   strcpy(dns[k].addr,addr);
   //  Look up hostname
   struct hostent* he = gethostbyaddr((char*)&tr->ip,sizeof(in_addr_t),AF_INET);
   if (he)
   {
      tr->fqdn = dns[k].fqdn = malloc(strlen(he->h_name)+1);
      strcpy(dns[k].fqdn,he->h_name);
   }
   else
      tr->fqdn = dns[k].fqdn = dns[k].addr;
   //  Return length
   return strlen(dns[k].fqdn);
}

//
//  Initialize traceroute
//
void InitTrace()
{
   tseq=0;
   tstat=0;
   for (int k=0;k<tTTL;k++)
   {
      tt[k].min = -1;
      tt[k].avg = -1;
      tt[k].max = -1;
      for (int i=0;i<nsec;i++)
         tt[k].ping[i] = 0xFF;
   }
}

//
//  Read data file
//
void InitPing(char* file[],const int nfile)
{
   char line[1024];
   int maxn=0;

   //  Initialize
   ntar = 0;
   nhdr = 0;
   seq  = 0;
   nwid = 6;  //  Minimum width
   awid = 6;  //  Addres width
   //  Initialize Traceroute array (max number of entries it tTTL)
   tt = (Trace*)malloc(tTTL*sizeof(Trace));
   //  Open first file in the list that is readable
   FILE* f=0;
   for (int k=0;k<nfile && !f;k++)
      f = fopen(file[k],"r");
   //  Cannot open any of the files
   if (!f)
   {
      fprintf(stderr,"Cannot open file %s",file[0]);
      for (int k=1;k<nfile;k++)
         fprintf(stderr," or %s",file[k]);
      Fatal("\n");
   }
   //  Read line
   int  indent=0;
   char* hdr=0;
   while (fgets(line,sizeof(line),f) != NULL )
   {
      char host[256];
      //  Skip comments
      if (line[0] == '#') continue;
      //  Trim trailing whitespace
      int l=strlen(line)-1;
      while (l>0 && isspace(line[l-1])) l--;
      line[l] = 0;
      //  Skip blank lines
      if (l==0) continue;
      //  Header line
      if (line[0] == '>')
      {
         //  Reset indent
         if (l==1 && indent>0)
            hdr = 0;
         //  Add header
         else
         {
            nhdr++;
            hdr = malloc(l+1);
            memcpy(hdr,line+1,l);
         }
         indent = l>1?3:0;
         continue;
      }
      //  Next target
      if (ntar+1>maxn)
      {
         maxn += 32;
         pt = (Target*)realloc(pt,maxn*sizeof(Target));
         if (!pt) Fatal("Out of memory allocating ping targets\n");
      }
      //  Set header
      pt[ntar].hdr = hdr;
      if (hdr) hdr = 0;
      //  Get hostname/ip and offset to start of display name
      int i;
     if (sscanf(line,"%255s %n",host,&i)!=1) Fatal("Error reading address: %s\n",line);
      //  Save display name and host name
      if (l>i)
      {
         //  Display name
         l = l-i+indent;
         pt[ntar].name = malloc(l+1);
         for (int j=0;j<indent;j++)
            pt[ntar].name[j] = ' ';
         memcpy(pt[ntar].name+indent,line+i,l-indent+1);
         if (l>nwid) nwid = l;
         //  Hostname
         l = strlen(host);
         pt[ntar].host = malloc(l+1);
         memcpy(pt[ntar].host,host,l+1);
         if (l>awid) awid = l;
      }
      //  Use host name as display name
      else
      {
         l = strlen(host)+indent;
         pt[ntar].name = malloc(l+1);
         for (int j=0;j<indent;j++)
            pt[ntar].name[j] = ' ';
         memcpy(pt[ntar].name+indent,host,l+1-indent);
         if (l>nwid) nwid = l;
         //  Blank host name
         pt[ntar].host = malloc(1);
         pt[ntar].host[0] = 0;
      }
      //  Initialize pings
      pstat = 0;
      for (i=0;i<nsec;i++)
         pt[ntar].ping[i] = 0xFF;
      pt[ntar].dt   = -1;
      pt[ntar].min  = -1;
      pt[ntar].avg  = -1;
      pt[ntar].max  = -1;
      //  Get IP address
      struct hostent* he = gethostbyname(host);
      if (!he) Fatal("Cannot resolve host name %s\n",host);
      memcpy(&pt[ntar].ip.s_addr,he->h_addr_list[0],4);
      //  Set up sa as an IP socket
      memset(&pt[ntar].sa,0,sizeof(struct sockaddr));
      struct sockaddr_in *isa = (struct sockaddr_in*)&pt[ntar].sa;
      isa->sin_family = AF_INET;
      isa->sin_addr.s_addr = pt[ntar].ip.s_addr;
      //  Check for duplicate IP addresses
      for (i=0;i<ntar;i++)
         if (pt[ntar].ip.s_addr==pt[i].ip.s_addr)
           Fatal("%s has a duplicate IP\n",pt[ntar].name);
      //  Increment ntar
      ntar++;
   }
   fclose (f);
   if (!ntar) Fatal("No tragets in %s\n",file);
}

//
//  Calculate checksum
//
unsigned short checksum(char* data,int len)
{
   int sum = 0;
   unsigned short csum = 0;
   unsigned short* word = (unsigned short*) data;
   while (len>1)
   {
     sum += *word++;
     len -= 2;
   }
   if (len==1)
   {
     *(unsigned char *)(&csum) = *(unsigned char *)word;
     sum += csum;
   }
   sum = (sum >> 16) + (sum & 0xffff);  // add hi 16 to low 16
   sum += (sum >> 16);                  // add carry
   csum = ~sum;                         // truncate to 16 bits
   return csum;
}

//
//  Draw symbol
//
void DrawSym(uint8_t num)
{
   if (num>9)
   {
      attron(A_BOLD);
      addch('X');
      attroff(A_BOLD);
   }
   else if (ich==3)
   {
      attron(A_BOLD);
      addch(num+'0');
      attroff(A_BOLD);
   }
   else if (pch)
      addch(pch);
   else if (ich==2)
      addch('*');
   else if (ich==1)
      addch(ACS_BLOCK);
   else
      addch(ACS_DIAMOND);
}

//
//  Draw ping
//
void DrawPing(int l,uint8_t ping)
{
   if (l<=0)
      addch('-');
   else if (ping==0xFF)
   {
      attron(COLOR_PAIR(5));
      DrawSym(0xF);
   }
   else
   {
      attron(COLOR_PAIR((ping&0xF)+2));
      DrawSym(ping>>4);
   }
}

//  Figure out the bottom row
int Bottom(int top)
{
   int i = (ntar+nhdr+1<hgt) ? 2 : 1;
   for (int k=top;k<ntar;k++)
   {
      i += pt[k].hdr ? 2 : 1;
      if (i==hgt)
         return k;
      else if (i>hgt)
         return k-1;
   }
   return ntar-1;
}

//
//  Print history header
//
void PrintHist(int n)
{
   //  Ping history header right to left
   if (r2l)
      for (int l=n-1;l>=0;l--)
      {
         if (l%10==0)
            addch('0');
         else if (l>10 && (l-1)%10==0)
            addch('0'+(l%100)/10);
         else if (l>100 && (l-2)%10==0)
            addch('0'+l/100);
         else
            addch(' ');
      }
   //  Ping history header left to right
   else
      for (int l=0;l<n;l++)
      {
         if (l%10==0)
            addch('0');
         else if (l>8 && (l+1)%10==0)
            addch('0'+((l+1)%100)/10);
         else if (l>97 && (l+2)%10==0)
            addch('0'+(l+2)/100);
         else
            addch(' ');
      }
}

//
//  Display
//
void Display()
{
   int bell = 0;
   //  Clear
   erase();
#ifdef piGPIO
   //  Do full refresh
   if (seq%10==0) clearok(curscr,1);
#endif
   attron(COLOR_PAIR(1));
   //  Help
   if (mode<0)
   {
      attron(A_BOLD);
      printw(" Key   Function\n");
      attroff(A_BOLD);
      printw(help);
   }
   //  Traceroute
   else if (mode)
   {
      //  Unwind trailing lack of response
      while (nhop>1 && !tt[nhop-1].ip && !tt[nhop-2].ip)
         nhop--;
      //  Display
      if (nhop+3<hgt) timeprint();
      attron(A_BOLD);
      printw("Traceroute to %s\n\n",pt[sel].name);
      //  Look up hostname and figure longest name
      int len=5;
      int lan=4;
      for (int k=0;k<nhop;k++)
      {
         int l = nslookup(tt+k);
         if (l>len) len = l;
         l = strlen(tt[k].addr);
         if (l>lan) lan = l;
      }
      //  Truncate hostnames if too long
      if (len+lan+12>wid) len = wid-12-lan;
      int ntrac = wid-13-len-lan;
      if (stat) ntrac -= 18;
      if (ntrac>nsec) ntrac = nsec;
      //  Print header
      printw("Hop Host");
      for (int k=0;k<len;k++)
         addch(' ');
      printw(" IP");
      for (int k=5;k<lan;k++)
         addch(' ');
      PrintHist(ntrac);
      printw("    ms");
      if (stat) printw("   min   avg   max");
      printw("\n");
      attroff(A_BOLD);
      int m = (nhop<hgt-3) ? nhop : hgt-3;
      //  Print replies
      for (int k=0;k<m;k++)
      {
         attron(COLOR_PAIR(1));
         printw("%3d ",k+1);
         char* ch = tt[k].fqdn;
         for (int l=0;l<len+1;l++)
            addch(*ch?*ch++:' ');
         ch = tt[k].addr;
         for (int l=0;l<lan+1;l++)
            addch(*ch?*ch++:' ');
         //  Pings
         if (r2l)
            for (int l=ntrac-1;l>=0;l--)
               DrawPing(tseq-l,tt[k].ping[l]);
         else
            for (int l=0;l<ntrac;l++)
               DrawPing(tseq-l,tt[k].ping[l]);
         //  Draw stats
         attron(COLOR_PAIR(1));
         if (tt[k].dt<0)
            printw(" unrch");
         else
            printw(" %5.1f",tt[k].dt);
         if (stat) printw(" %5.1f %5.1f %5.1f",tt[k].min,tt[k].avg,tt[k].max);
         printw("\n");
         bell = bell | (tt[k].dt<0);
      }
   }
   //  Ping
   else
   {
      int i=1;
      //  Print header if we have enough rows
      if (ntar+nhdr+1<hgt)
      {
         timeprint();
         i++;
      }
      //
      //  Print header line
      //
      attron(A_BOLD);
      //  Display name
      printw("Target");
      for (int l=6;l<nwid;l++)
         addch(' ');
      //  Host/IP if selected
      if (showip)
      {
         printw(" Address");
         for (int l=7;l<awid;l++)
            addch(' ');
      }
      PrintHist(nping);
      //  Ping times
      printw("   ms");
      //  Number of hops
      if (hop) printw(" hop");
      //  Stats
      if (stat) printw("   min   avg   max");
      attroff(A_BOLD);
      //
      //  Draw ping table
      //
      for (int k=top;k<ntar;k++)
      {
         //  Bail out at the bottom of the screen
         if (i>=hgt) break;
         //  Target name
         move(i++,0);
         if (pt[k].hdr)
         {
            attron(A_BOLD);
            printw(pt[k].hdr);
            attroff(A_BOLD);
            move(i++,0);
            //  Bail out at the bottom of the screen
            if (i>=hgt) break;
         }
         //  Choose color
         if (k==sel)
            attron(COLOR_PAIR(4));
         else
            attron(COLOR_PAIR(1));
         //  Print name
         char* ch = pt[k].name;
         for (int l=0;l<nwid;l++)
            addch(*ch?*ch++:'.');
         //  Print address
         if (showip)
         {
            addch(' ');
            char* ch = pt[k].host;
            for (int l=0;l<awid;l++)
               addch(*ch?*ch++:'.');
         }
         if (k==sel) attron(COLOR_PAIR(1));
         //  Pings
         if (r2l)
            for (int l=nping-1;l>=0;l--)
               DrawPing(seq-l,pt[k].ping[l]);
         else
            for (int l=0;l<nping;l++)
               DrawPing(seq-l,pt[k].ping[l]);
         //  Ping time
         attron(COLOR_PAIR(1));
         if (pt[k].dt<0)
            printw(" -----");
         else
            printw(" %5.1f",pt[k].dt);
         //  Hop count
         int l = pTTL+1-pt[k].ttl;
         if (!hop)
         {}
         else if (pt[k].dt<0 || l<0)
            printw(" --");
         else
            printw(" %2d",l);
         //  Draw stats
         if (stat) printw(" %5.1f %5.1f %5.1f",pt[k].min,pt[k].avg,pt[k].max);
         //  Bell if pings
         bell = bell | (tt[k].dt<0) | (seq>1 && pt[k].ping[0]==0xFF);
      }
   }
   show = 0;
   if (bell && !silent) beep();
   refresh();
}

//
//  Send ICMP packet
//
void ICMP(int id,int seq,int ttl,struct sockaddr sa)
{
   char buf[8192];
   //  Set TTL
   if (setsockopt (sock,SOL_IP,IP_TTL,&ttl,sizeof(ttl))<0) Fatal("Cannot set TTL\n");
   //  Set up ICMP packet header
   struct icmphdr* icp = (struct icmphdr*)buf;
   icp->type       = ICMP_ECHO;
   icp->code       = 0;
   icp->checksum   = 0;
   icp->un.echo.id       = id;
   icp->un.echo.sequence = seq;
   //  Set up the payload
   double time = now();
   int len = sizeof(struct icmphdr);
   memcpy(buf+len,&time,sizeof(double));
   len += sizeof(double);
   //  Compute checksum
   icp->checksum = checksum(buf,len);
   //  Send packet
   int i = sendto(sock,buf,len,0,&sa,sizeof(struct sockaddr));
#ifdef __APPLE__
   //  Some OSX machines inexplicably return -1
   if (i>0 && i!=len) fprintf(stderr,"Failed to send ICMP packet\n");
#else
   if (i<0 || i!=len) fprintf(stderr,"Failed to send ICMP packet\n");
#endif
}

//
//  Shift ping buffer
//
void PingShift(uint8_t ping[])
{
   for (int i=nsec-1;i>0;i--)
      ping[i] = ping[i-1];
   ping[0] = 0xFF;
}

//
//  Send ping to all targets
//
void* Ping()
{
   while (1)
   {
      //  Parallel traceroute
      tseq++;
      tstat++;
      if (tseq>65535) tseq=nsec;
      nhop = tTTL;
      for (int k=0;k<nhop;k++)
      {
         //  Initialize trace
         tt[k].dt = 0;
         tt[k].ip = 0;
         PingShift(tt[k].ping);
         //  Send Ping
         ICMP(traceid,k+1,k+1,pt[sel].sa);
         //  Pause before sending next
         usleep(pus);
      }
      //  Ping all targets with TTL pTTL
      seq++;
      pstat++;
      if (seq>65535) seq=nsec;
      for (int k=0;k<ntar;k++)
      {
         //  Advance ping array
         //  This should be a ring buffer, but I'm too lazy
         PingShift(pt[k].ping);
         // Send Ping
         ICMP(pingid,seq,pTTL,pt[k].sa);
         //  Pause before sending next
         usleep(pus);
      }
      //  Pause until next second
      usleep(950000-ntar*pus);
      show = 1;
      //  Give display 50ms to update
      usleep((sbp-1)*1000000+50000);
   }
}

//
//  Unpack IP/ICMP header
//
int UnpackHeader(unsigned char* data,int l,int* ttl,int* rtp, int* rid,int* rsq)
{
   int len=0;
   //  Get TTL
   if (l<sizeof(struct ip)) return 0;
   struct ip* ip = (struct ip*)data;
   *ttl = ip->ip_ttl;
   //  Skip the IP header
   int hlen = ip->ip_hl << 2;
   data += hlen;
   len  += hlen;
   l    -= hlen;
   //  Unpack ICMP header
   if (l<sizeof(struct icmphdr)) return 0;
   struct icmphdr* icp = (struct icmphdr*)data;
   *rtp = icp->type;
   *rid = icp->un.echo.id;
   *rsq = icp->un.echo.sequence;
   data += sizeof(struct icmphdr);
   len  += sizeof(struct icmphdr);
   l    -= sizeof(struct icmphdr);
   return len;
}

//
//  Encode time as byte
//  Upper nibble 0-9 lower nibble exponent base 10
//  Example 0x52 = 500
//
uint8_t ByteTime(double dt)
{
   int idt = dt+0.5;
   uint8_t bdt=0xFF;
   if (idt<10)
      bdt = (idt<<4) + 0;
   else if (idt<100)
      bdt = ((idt/10)<<4) + 1;
   else if (idt<1000)
      bdt = ((idt/100)<<4) + 2;
   else if (idt<10000)
      bdt = ((idt/1000)<<4) + 3;
   return bdt;
}

//
//  Update ping stats
//
void Stats(double dt,int n,double* min,double* avg,double* max)
{
   if (*min<0 || dt<*min) *min = dt;
   if (*max<0 || dt>*max) *max = dt;
   if (*avg<0)
      *avg = dt;
   else
      *avg = (*avg*(n-1)+dt)/n;
}

//
//  Receive pings
//
void* Receive()
{
   while (1)
   {
      unsigned char buf[8192];
      unsigned char* data=buf;
      //  Check for reply
      struct sockaddr from;
      unsigned int flen = sizeof(from);
      int l = recvfrom(sock,buf,8192,0,(struct sockaddr *)&from,&flen);
      if (l<0) continue;
      //  Check if this packet is from a known host
      struct sockaddr_in* isa = (struct sockaddr_in*)&from;
      int host = -1;
      for (int i=0;i<ntar && host<0;i++)
         if (isa->sin_addr.s_addr==pt[i].ip.s_addr) host = i;
      //  Unpack header
      int ttl,rtp,rid,rsq;
      int off = UnpackHeader(data,l,&ttl,&rtp,&rid,&rsq);
      if (!off) continue;
      data += off;
      l    -= off;
      //  Process echo reply
      if (rtp==ICMP_ECHOREPLY && l>=sizeof(double))
      {
         //  Calculate delay
         double t0 = *(double*)data;
         double dt = 1000*(now()-t0);
         //  Current ping reply from known host
         if (rid==pingid && rsq==seq && host>=0)
         {
            pt[host].ttl = ttl;
            pt[host].dt  = dt;
            pt[host].ping[0] = ByteTime(dt);
            Stats(dt,pstat,&pt[host].min,&pt[host].avg,&pt[host].max);
         }
         //  Traceroute reply
         else if (rid==traceid && rsq>0 && rsq<=nhop)
         {
            //  Length of path
            if (rsq<nhop) nhop = rsq;
            tt[rsq-1].dt = dt;
            tt[rsq-1].ip = isa->sin_addr.s_addr;
            tt[rsq-1].ping[0] = ByteTime(dt);
            Stats(dt,tstat,&tt[rsq-1].min,&tt[rsq-1].avg,&tt[rsq-1].max);
         }
      }
      //  Traceroute time exceeded
      else if (rtp==ICMP_TIME_EXCEEDED)
      {
         //  Data is original packet
         off = UnpackHeader(data,l,&ttl,&rtp,&rid,&rsq);
         if (!off) continue;
         data += off;
         l    -= off;
         if (l<sizeof(double)) continue;
         double t0 = *(double*)data;
         double dt = 1000*(now()-t0);
         if (rid==traceid && rsq>0 && rsq<=nhop)
         {
            tt[rsq-1].dt = dt;
            tt[rsq-1].ip = isa->sin_addr.s_addr;
            tt[rsq-1].ping[0] = ByteTime(dt);
            Stats(dt,tstat,&tt[rsq-1].min,&tt[rsq-1].avg,&tt[rsq-1].max);
         }
      }
      //  Destination unreachable
      else if (rtp==ICMP_UNREACH)
      {
         //  Data is original packet
         off = UnpackHeader(data,l,&ttl,&rtp,&rid,&rsq);
         if (!off) continue;
         if (rid==traceid && rsq>0 && rsq<nhop)
         {
            nhop = rsq;
            tt[rsq-1].dt = -1;
            tt[rsq-1].ip = isa->sin_addr.s_addr;
         }
      }
   }
}

//  Select new target
void newsel(int new)
{
   //  Reset traceroute hops when selected target changes
   nhop = 0;
   InitTrace();
   //  In ping mode stick to the selected screen
   int min = mode ? 0 : top;
   int max = mode ? ntar-1 : Bottom(top);
   if (new>max)
     sel = min;
   else if (new<min)
     sel = max;
   else
     sel = new;
}

//  Check that selected item is in range
//  dir=0 is resizing the screen
void Scroll(int dir)
{
   //  Everything fits
   if (ntar+nhdr+1<hgt)
      top = 0;
   //  Ping mode scroll or resize
   else if (!mode)
   {
      //  Current bottom
      int bot = Bottom(top);
      // Scroll down
      if (dir>0)
      {
         for (int i=0;i<dir && Bottom(top)<ntar-1;i++)
            bot = Bottom(++top);
      }
      //  Scroll up
      else if (dir<0)
      {
         top = top+dir>0 ? top+dir : 0;
         bot = Bottom(top);
      }
      //  Check that selection is on the screen
      if (sel<top) sel = top;
      if (sel>bot) sel = bot;
   }
   //  Resize in traceroute mode
   else if (!dir)
   {
      //  Current bottom
      int bot = Bottom(top);
      //  Make sure that selection remains on the screen
      if (sel<top) top = sel;
      if (sel>bot) top -= sel-bot;
      if (top<0) top = 0;
   }
}

//
//  Set window size
//
void Resize()
{
   int nx = hop ? nwid+9 : nwid+6;
   if (showip) nx += awid + 1;
   getmaxyx(stdscr,hgt,wid);
   Scroll(0);
   nping = wid - nx;
   if (stat) nping -= 18;
   if (nping>nsec) nping = nsec;
}

//
//  Set color
//
void SetColor()
{
   if (white)
   {
      init_pair(1,COLOR_BLACK,COLOR_WHITE);
      init_pair(2,COLOR_CYAN,COLOR_WHITE);
      init_pair(3,COLOR_GREEN,COLOR_WHITE);
      init_pair(4,COLOR_YELLOW,COLOR_WHITE);
      init_pair(5,COLOR_RED,COLOR_WHITE);
   }
   else
   {
      init_pair(1,COLOR_WHITE,COLOR_BLACK);
      init_pair(2,COLOR_CYAN,COLOR_BLACK);
      init_pair(3,COLOR_GREEN,COLOR_BLACK);
      init_pair(4,COLOR_YELLOW,COLOR_BLACK);
      init_pair(5,COLOR_RED,COLOR_BLACK);
   }
   bkgd(COLOR_PAIR(1));
}

//
//  Initialize curses
//
void InitCurses()
{
   // Start curses with raw input
   initscr();
   nonl(); cbreak(); noecho(); nodelay(stdscr,TRUE); keypad(stdscr,TRUE); curs_set(0);
   //  Set colors
   if (!has_colors()) FatalW("No color support\n");
   start_color();
   init_color(COLOR_WHITE,1000,1000,1000);
   SetColor();
   // Set window size
   Resize();
}

//
//  Initialize wiringPi
//
#ifdef piGPIO
void gpio(int pin,int level,uint32_t tick)
{
   //  Ignore rapid interrupts (probably contact bounce)
   double t = now();
   if (t-swt<0.3) return;
   //  Check for mapped pin
   for (int k=0;k<4;k++)
     if (pin==SW[k]) swx = k+1;
   //  Remember time
   if (swx) swt = t;
}
void InitPIgpio(void)
{
   //  Initialize piGPIO
   if (gpioInitialise()<0) FatalW("Cannot initialize GPIO\n");
   //  Set pins for input, pull-up and interrupt
   for (int k=0;k<4;k++)
   {
      if (gpioSetMode(SW[k],PI_INPUT)) FatalW("Error setting SW%d to input\n",SW[k]);
      if (gpioSetPullUpDown(SW[k],PI_PUD_UP)) FatalW("Error setting SW%d to pull-up\n",SW[k]);
      if (gpioSetISRFunc(SW[k],FALLING_EDGE,-1,&gpio)) FatalW("Error setting interrupt for SW%d\n",SW[k]);
   }
}
#endif

//
//  Init ICMP socket
//
void InitSock(int init)
{
   //  Close old socket
   if (!init) close(sock);

   //  Get unique IDs for ping and traceroute
   pingid = (getpid() & 0x7FFF) << 1;
   traceid = pingid|0x01;

   struct protoent* proto = getprotobyname("icmp");
   if (!proto) FatalW("icmp protocol not defined\n");

   //  Set up socket
   sock = socket(AF_INET,SOCK_RAW,proto->p_proto);
   if (sock<0) FatalW("Cannot open ICMP socket\n");

   //  Show reset
   move(0,0);
   printw("********RESET*******");
}

int main(int argc,char* argv[])
{
   //
   //  Process command line arguments
   //
   int ch;
   int   nfile = 2;
   char* file[2] = {"cping.cfg","/etc/cping.cfg"};
   while ((ch = getopt(argc,argv,"banrgxths:p:f:c:")) != EOF)
   {
       //  Black background
       if (ch == 'b')
          white = 0;
       //  Do not show hops
       else if (ch == 'n')
          hop = 0;
       //  Reverse direction
       else if (ch == 'r')
          r2l = 0;
       //  Microseconds between pings
       else if (ch == 'p')
          pus = atoi(optarg);
       //  Show address
       else if (ch == 'a')
          showip = 1;
       //  Show numbers
       else if (ch == 'x')
          ich = 3;
       //  Show stats
       else if (ch == 't')
          stat = 1;
       //  Seconds beteen ping groups
       else if (ch == 's')
       {
          sbp = atoi(optarg);
          if (sbp<1 || sbp>5) Fatal("Invalid -s %d\n",sbp);
       }
       //  Ping character
       else if (ch == 'c')
          pch = optarg[0];
       //  File
       else if (ch == 'f')
       {
          nfile = 1;
          file[0] = optarg;
       }
       //  Help
       else if (ch == 'h')
          Fatal("Usage: cping [-banrgxth] [-p us] [-f file]\n" 
                "  -b  White lettering on black background\n"
                "  -a  Show address in ping table\n"
                "  -n  No hops on ping table\n"
                "  -r  Scroll pings left to right\n"
                "  -p  microseconds between pings [default 1000]\n"
                "  -f  config file [default cping.cfg or /etc/cping.cfg]\n"
                "  -s  seconds between ping\n"
                "  -x  show numeric ping character\n"
                "  -t  show ping time stats\n"
                "  -h  help\n\n"
		" Ping targets are read from cping.cfg\n"
		" Each line is a target name followed by an ip address\n"
		" ~ in the target name becomes a space\n"
		" # in column 1 starts a comment\n\n"
                " Key  Function\n%s",help);
#ifdef piGPIO
       //  Initialize wiringPi
       else if (ch == 'g')
          InitPIgpio();
#endif
       else
          Fatal("Unknown option %c\n",ch);
   }
   //  Read data
   InitPing(file,nfile);
   if (pus*(ntar+tTTL)>950000) Fatal("Pause length exceeds one second\n");
   //  Initialize curses
   InitCurses();
   //  Initialize ICMP socket
   InitSock(1);
   //  Initialize DNS
   InitDNS();
   //  Start read thread
   if (pthread_create(&rd,NULL,Receive,NULL)) FatalW("Cannot start receive thread\n");
   //  Start write thread
   if (pthread_create(&rd,NULL,Ping,NULL)) FatalW("Cannot start write thread\n");
   //  Main loop
   for (int run=1;run;)
   {
      int ch = getch();
      //  Quit
      if (ch=='q')
         run = 0;
      //  Window resized
      else if (ch==KEY_RESIZE)
      {
         Resize();
         Display();
      }
      //  Scroll down
      else if (ch==KEY_DOWN)
      {
         Scroll(+1);
         Display();
      }
      //  Scroll down
      else if (ch==KEY_NPAGE)
      {
         Scroll(+hgt);
         Display();
      }
      //  Scroll up
      else if (ch==KEY_UP)
      {
         Scroll(-1);
         Display();
      }
      //  Scroll up
      else if (ch==KEY_PPAGE)
      {
         Scroll(-hgt);
         Display();
      }
      //  Select previous target for traceroute
      else if (ch==KEY_LEFT)
      {
         newsel(sel-1);
         Display();
      }
      //  Select next target for traceroute
      else if (ch==KEY_RIGHT)
      {
         newsel(sel+1);
         Display();
      }
      //  Toggle mode
      else if (swx==1)
      {
         swx = 0;
         mode = !mode;
         Display();
      }
      //  Previous target with scroll
      else if (swx==2)
      {
         swx = 0;
         if (sel<=top) Scroll(-1);
         newsel(sel-1);
         Display();
      }
      //  Next target with scroll
      else if (swx==3)
      {
         swx = 0;
         if (sel>=Bottom(top)) Scroll(+1);
         newsel(sel+1);
         Display();
      }
      //  Toggle display of addresses
      else if (swx==4)
      {
         swx = 0;
         showip = !showip;
         Resize();
         Display();
      }
      //  Switch to traceroute mode
      else if (ch==KEY_ENTER || ch=='\n' || ch=='\r')
      {
         mode = !mode;
         Display();
      }
      //  Switch to ping mode
      else if (ch==27) // ESC
      {
         mode = 0;
         Display();
      }
      //  Toggle hops in display
      else if (ch=='n')
      {
         hop = 1-hop;
         Resize();
         Display();
      }
      //  Invert colors
      else if (ch=='i')
      {
         white = !white;
         SetColor();
         Display();
      }
      //  Reverse ping display direction
      else if (ch=='r')
      {
         r2l = !r2l;
         Display();
      }
      //  Toggle display of addresses
      else if (ch=='a')
      {
         showip = !showip;
         Resize();
         Display();
      }
      //  Toggle stats
      else if (ch=='t')
      {
         stat = !stat;
         Resize();
         Display();
      }
      //  Toggle silent
      else if (ch=='s')
         silent = !silent;
      //  Show help
      else if (ch=='h')
         mode = -1;
      //  Toggle ping character
      else if (ch=='c')
         ich = (ich+1)%4;
      //  Reset ping socket
      else if (ch=='0')
         InitSock(0);
      //  Update display
      else if (show)
         Display();
      //  Sleep 1ms
      usleep(1000);
   }
   endwin();
#ifdef piGPIO
   gpioTerminate();
#endif
   return 0;
}
